{{
************************************
* File jvmInvoke: invoking methods *
************************************
}}

{{
  Invoking Methods. 
  Method invocation is a complex process. To invoke a method, the operand stack must contain a reference 
  to an object and some number of arguments. The object reference is used as a pointer to the object's 
  method table, which contains the method signature. The method signature is guaranteed to exactly match 
  one of the method signatures in the table. The arguments byte1 and byte2 are used to construct an index 
  into the constant pool of the current class, which contains the complete method signature.  
  The result of the lookup is an index into the method table of the named class. This index is used with the 
  referenced object's dynamic type to look in the method table of that type, where a pointer to the method 
  block for the matched method is found. The method block indicates the type of method, such as native or 
  synchronized, and the number of arguments expected on the operand stack.  
  The object reference and arguments are popped off the method's operand stack and become the initial 
  values of the local variables of the new method. Execution continues with the first instruction of the new 
  method. 
  Note:
  The monitor associated with the referenced object is entered if the method is marked as synchronized.
  A NullPointerException is thrown if the object reference on the operand stack is null. A StackOverflowError
  is thrown if a stack overflow is detected during the method invocation.
}}

{{
  invokevirtual
  Operation: Invoke instance method; dispatch based on class
  Format: invokevirtual,indexbyte1,indexbyte2   
  Operand Stack: ...,objectref,[arg1,[arg2 ...]] --> ...,

  Description
  The unsigned indexbyte1 and indexbyte2 are used to construct an index into the runtime constant pool of the
  current class, where the value of the index is (indexbyte1 << 8) | indexbyte2. The runtime constant pool item
  at that index must be a symbolic reference to a method, which gives the name and descriptor of the method as
  well as a symbolic reference to the class in which the method is to be found. The named method is resolved.
  The method must not be an instance initialization method or the class or interface initialization method.
  Finally, if the resolved method is protected, and it is either a member of the current class or a member of
  a superclass of the current class, then the class of objectref must be either the current class or a subclass
  of the current class.
  Let C be the class of objectref. The actual method to be invoked is selected by the following lookup procedure:
  If C contains a declaration for an instance method with the same name and descriptor as the resolved method,
  and the resolved method is accessible from C, then this is the method to be invoked, and the lookup procedure
  terminates. Otherwise, if C has a superclass, this same lookup procedure is performed recursively using the
  direct superclass of C ; the method to be invoked is the result of the recursive invocation of this lookup
  procedure. Otherwise, an AbstractMethodError is raised. 
  The objectref must be followed on the operand stack by nargs argument values, where the number, type, and order
  of the values must be consistent with the descriptor of the selected instance method.
  If the method is synchronized, the monitor associated with objectref is acquired or reentered.
  If the method is not native, the nargs argument values and objectref are popped from the operand stack.
  A new frame is created on the Java virtual machine stack for the method being invoked. The objectref and the
  argument values are consecutively made the values of local variables of the new frame, with objectref in local
  variable 0, arg1 in local variable 1 (or, if arg1 is of type long or double, in local variables 1 and 2),
  and so on. Any argument value that is of a floating-point type undergoes value set conversion prior to being
  stored in a local variable. The new frame is then made current, and the Java virtual machine pc is set to the
  opcode of the first instruction of the method to be invoked. Execution continues with the first instruction
  of the method.
  If the method is native and the platform-dependent code that implements it has not yet been bound into the Java
  virtual machine, that is done. The nargs argument values and objectref are popped from the operand stack and are
  passed as parameters to the code that implements the method. Any argument value that is of a floating-point type
  undergoes value set conversion prior to being passed as a parameter. The parameters are passed and the code is
  invoked in an implementation-dependent manner. When the platform-dependent code returns, the following take place:
  If the native method is synchronized, the monitor associated with objectref is released or exited as if by
  execution of a monitorexit instruction. If the native method returns a value, the return value of the
  platform-dependent code is converted in an implementation-dependent way to the return type of the native method
  and pushed onto the operand stack. 
  Runtime Exceptions
  If objectref is null, the invokevirtual instruction throws a NullPointerException. 
  Notes
  The nargs argument values and objectref are not one-to-one with the first nargs + 1 local variables.
  Argument values of types long and double must be stored in two consecutive local variables, thus more
  than nargs local variables may be required to pass nargs argument values to the invoked method.
}}

{{
  invokestatic
  Operation: Invoke a class (static) method
  Format: invokestatic,indexbyte1,indexbyte2   
  Operand Stack: ...,[arg1,[arg2 ...]] --> ...,

  Description
  The unsigned indexbyte1 and indexbyte2 are used to construct an index into the runtime constant pool of the
  current class, where the value of the index is (indexbyte1 << 8) | indexbyte2. The runtime constant pool item
  at that index must be a symbolic reference to a method, which gives the name and descriptor of the method as
  well as a symbolic reference to the class in which the method is to be found. The named method is resolved.
  The method must not be the class or interface initialization method. It must be static, and therefore cannot
  be abstract. 
  On successful resolution of the method, the class that declared the resolved field is initialized if that class
  has not already been initialized.
  The operand stack must contain nargs argument values, where the number, type, and order of the values must be
  consistent with the descriptor of the resolved method.
  If the method is synchronized, the monitor associated with the resolved class is acquired or reentered.
  If the method is not native, the nargs argument values are popped from the operand stack. A new frame is
  created on the Java virtual machine stack for the method being invoked. The nargs argument values are
  consecutively made the values of local variables of the new frame, with arg1 in local variable 0 (or, if arg1
  is of type long or double, in local variables 0 and 1) and so on. Any argument value that is of a floating-point
  type undergoes value set conversion prior to being stored in a local variable. The new frame is then made current
  and the Java virtual machine pc is set to the opcode of the first instruction of the method to be invoked.
  Execution continues with the first instruction of the method.
  If the method is native and the platform-dependent code that implements it has not yet been bound into the Java
  virtual machine, that is done. The nargs argument values are popped from the operand stack and are passed as
  parameters to the code that implements the method. Any argument value that is of a floating-point type undergoes
  value set conversion prior to being passed as a parameter. The parameters are passed and the code is invoked in
  an implementation-dependent manner. When the platform-dependent code returns, the following take place:
  If the native method is synchronized, the monitor associated with the resolved class is released or exited as
  if by execution of a monitorexit instruction. If the native method returns a value, the return value of the
  platform-dependent code is converted in an implementation-dependent way to the return type of the native method
  and pushed onto the operand stack. 

  Notes
  The nargs argument values are not one-to-one with the first nargs local variables. Argument values of types long
  and double must be stored in two consecutive local variables, thus more than nargs local variables may be required
  to pass nargs argument values to the invoked method.
}}

{{
  stackframe just before invoke       stackframe just after invoke       stackframe just after return

       |             |                     |             |                    |             |
       +-------------+                     +-------------+ <- fp              +-------------+
       | aref / arg0 |                     | aref / arg0 |                    | return val  |
       +-------------+                     +-------------+                    +-------------+ <- sp
       | arg0 / arg1 |                     | arg0 / arg1 |
       |    .....    |                     |    .....    |
       +-------------+ <- sp               +-------------+
                                           |   locals    |
                                           |    .....    |
                                           +-------------+
                                           | old method  |
                                           +-------------+
                                           |   old fp    |
                                           +-------------+
                                           | return addr |
                                           +-------------+ <- sp
                                           
}}


OBJ
  js:   "jvmStack"
  jd:   "jvmData"
  comm: "jvmComm"
  jn:   "jvmNative"
  jsub: "jvmSubReturn"
  jo:   "jvmObject"
  je:   "jvmException"
  
  
PUB jb_invokevirtual | aref,sp,fp,class,method,nf,len,str,pc
{{ Meaning: Invoke method based on runtime type. }}
{{
  JEM format: invokevirtual,byte1,byte2
  byte2 is offset to begin of class method table (so if byte = $2D and class method table = $60, method header = $8D)
  byte1 is offset to sp to find object reference (so if sp = $3FF6 and byte1 = $04 then the aref is at $3FF6-$04 = $3FF2)
}}
  sp := jd.getJvmSP                                       'get sp
  aref := jd.readJavaByte(jd.getJvmPC+1)                  'get offset from sp where aref is stored
  fp := sp - aref                                         'calculate address of possibly new frame
  aref := jd.readJavaIntBE(fp)                            'get aref
  if aref <> 0
    class := jd.readJavaIntBE(aref)                         'get address of class descriptor
    method := jd.readJavaByte(jd.getJvmPC+2) + class        'get address of method descriptor
    if jd.readJavaByte(method + jd#JVM_FLAGS) == $00        'not native
      sp += jd.readJavaByte(method + jd#JVM_LOCALS)         'skip locals     
      jd.writeJavaIntBE(sp,jd.getJvmMP)                     'save old mp
      jd.writeJavaIntBE(sp+2,jd.getJvmFP)                   'save old fp
      jd.writeJavaIntBE(sp+4,jd.getJvmPC+3)                 'save old pc (= return address)
      jd.setJvmFP(fp)                                       'set new fp
      jd.setJvmMP(method)                                   'set new mp
      jd.setJvmSP(sp+6)                                     'set new sp     
      jd.setJvmPC(jd.readJavaIntBE(method + jd#JVM_CODE))   'set new pc
    if jd.readJavaByte(method + jd#JVM_FLAGS) == $80        'native
      nf := jd.readJavaIntBE(method + jd#JVM_CODE)          'get native function number
      jn.doNative(nf)
      jd.incJvmPC(3)
  else
    je.throw(jd#kNullPointerException)

    
{{PUB jb_invokenonvirtual}}
{{ Meaning: Invoke method based on compile-time type. }}


PUB jb_invokestatic | sp,fp,method,nf,len,str
{{ Meaning: Invoke a class (static) method. }}
{{
  JEM format: invokestatic,byte1,byte2
  byte1 and byte2 form 16bits address to method header: addr = (byte1<<8)|(byte2) 
}}
  sp := jd.getJvmSP                                       'get sp
  method := jd.readJavaIntBE(jd.getJvmPC+1)               'get address of method descriptor
  fp := sp - (jd.readJavaByte(method + jd#JVM_PARAMETERS) & $7F)  'calculate address of possible new frame (mask out exception bit)
  if jd.readJavaByte(method + jd#JVM_FLAGS) == $00        'not native
    sp += jd.readJavaByte(method + jd#JVM_LOCALS)         'skip locals     
    jd.writeJavaIntBE(sp,jd.getJvmMP)                     'save old mp
    jd.writeJavaIntBE(sp+2,jd.getJvmFP)                   'save old fp
    jd.writeJavaIntBE(sp+4,jd.getJvmPC+3)                 'save old pc (= return address)
    jd.setJvmFP(fp)                                       'set new fp
    jd.setJvmMP(method)                                   'set new mp
    jd.setJvmSP(sp+6)                                     'set new sp     
    jd.setJvmPC(jd.readJavaIntBE(method + jd#JVM_CODE))   'set new pc
  if jd.readJavaByte(method + jd#JVM_FLAGS) == $80        'native
    nf := jd.readJavaIntBE(method + jd#JVM_CODE)          'get native function number
    jn.doNative(nf)
    jd.incJvmPC(3)
  

{{PUB jb_invokeinterface}}
{{ Meaning: Invoke interface method. }} 
{{
  Note:
  Unlike invokevirtual and invokenonvirtual, the method block does not indicate the number
  of available arguments. This number is taken from the bytecode.
}}
      
    